"""
This example demonstrates how to acquire a sample of a specified duration.
Sample rate is handled by the device and is very accurate.
Use this method to acquire high quality data for further analysis.
"""

import logging

import picoquake

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    # Create a PicoQuake device
    device = picoquake.PicoQuake("d4e9")

    # Configure acquisition
    device.configure(
        sample_rate=picoquake.SampleRate.hz_1000,
        filter_hz=picoquake.Filter.hz_213,
        acc_range=picoquake.AccRange.g_16,
        gyro_range=picoquake.GyroRange.dps_2000,
    )

    # Alternatively, configure from float values.
    # The closest available values will be used.
    device.configure_approx(
        sample_rate=1000,
        filter_hz=200,
        acc_range=16,
        gyro_range=2000,
    )

    # Acquire data from the device
    data, exception = device.acquire(seconds=1)

    # Stop the device
    device.stop()

    # Check if an exception occurred
    if exception is not None:
        raise exception

    # Print the acquired data
    print(f"Data: {data}")

    # Save to a CSV file
    data.to_csv("acquisition.csv")



from itertools import permutations
from collections import deque
from typing import List, Any, Optional
import os


def get_axis_combinations(axis: str) -> set:
    return set(''.join(p) for i in range(1, len(axis) + 1) for p in permutations(axis, i))


def deque_get_last_n(data: deque, n: int) -> List[Any]:
    """
    Get the last n elements from a deque. Does not remove the elements from the deque.

    Args:
    data: Deque with data.
    n: Number of elements to get.

    Returns:
    List with the last n elements from the deque. If n is greater than the length of the deque, all elements are returned.
    """
    start_idx = max(0, len(data) - n)
    return [data[i] for i in range(start_idx, len(data))]


def deque_slice(dq: deque, start: Optional[int], end: Optional[int] = None) -> List[Any]:
    """
    Return a slice from the deque. Behaves like the list slice method.
    
    Args:
    dq: The deque to slice.
    start: The starting index of the slice.
    end : The ending index of the slice.
    
    Returns:
    deque: A deque containing the specified slice.
    """
    if start is None:
        start = 0
    if end is None:
        end = len(dq)
    if start < 0:
        start = len(dq) + start
    if end < 0:
        end = len(dq) + end
    if start < 0:
        start = 0
    if end > len(dq):
        end = len(dq)
    return [dq[i] for i in range(start, end)]


def get_unique_filename(path: str):
    """
    Returns a unique filename by appending a sequential number if the file already exists.
    
    Args:
        path: The path to the original file.
    
    Returns:
        A unique filename.
    """
    if not os.path.isfile(path):
        return path
    
    base, extension = os.path.splitext(path)
    counter = 1
    new_path = f"{base}_{counter}{extension}"
    
    while os.path.isfile(new_path):
        counter += 1
        new_path = f"{base}_{counter}{extension}"
    
    return new_path



import os
from tempfile import TemporaryDirectory

from picoquake.utils import *


def test_get_axis_combinations():
    assert get_axis_combinations("x") == {"x"}
    assert get_axis_combinations("xyz") == {"x", "y", "z",
                                            "xy", "xz", "yx", "yz", "zx", "zy",
                                            "xyz", "xzy", "yxz", "yzx", "zxy", "zyx"}
    assert get_axis_combinations("abcd") == {"a", "b", "c", "d",
                                            "ab", "ac", "ad", "ba", "bc", "bd", "ca", "cb", "cd", "da", "db", "dc",
                                            "abc", "abd", "acb", "acd", "adb", "adc", "bac", "bad", "bca", "bcd", "bda", "bdc",
                                            "cab", "cad", "cba", "cbd", "cda", "cdb", "dab", "dac", "dba", "dbc", "dca", "dcb",
                                            "abcd", "abdc", "acbd", "acdb", "adbc", "adcb", "bacd", "badc", "bcad", "bcda", "bdac", "bdca",
                                            "cabd", "cadb", "cbad", "cbda", "cdab", "cdba", "dabc", "dacb", "dbac", "dbca", "dcab", "dcba"}


def test_deque_get_last_n():
    d = deque([1, 2, 3, 4, 5])
    assert deque_get_last_n(d, 3) == [3, 4, 5]
    assert deque_get_last_n(d, 5) == [1, 2, 3, 4, 5]
    assert deque_get_last_n(d, 6) == [1, 2, 3, 4, 5]
    assert deque_get_last_n(d, 0) == []
    assert deque_get_last_n(d, -1) == []
    d = deque()
    assert deque_get_last_n(d, 3) == []
    assert deque_get_last_n(d, 0) == []
    assert deque_get_last_n(d, -1) == []


def test_deque_slice():
    dq = deque([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    test_slices = [(None, None), (None, 3), (0, 3), (0, 10), (3, 7),
                   (None, -1), (-5, None), (-5, -3), (0, -3), (3, -3), (0, -10)]
    for start, end in test_slices:
        assert deque_slice(dq, start, end) == list(dq)[start:end]
    

def test_get_unique_filename():
    with TemporaryDirectory() as temp_dir:
        # Define a base filename
        base_filename = os.path.join(temp_dir, "file.csv")
        
        # Test when file does not exist
        assert get_unique_filename(base_filename) == base_filename
        
        # Create the base file
        with open(base_filename, 'w') as f:
            f.write('test content')

        # Test when file exists
        assert get_unique_filename(base_filename) == os.path.join(temp_dir, "file_1.csv")
        
        # Create the first unique file
        with open(os.path.join(temp_dir, "file_1.csv"), 'w') as f:
            f.write('test content')

        # Test when base file and first unique file exist
        assert get_unique_filename(base_filename) == os.path.join(temp_dir, "file_2.csv")



from pytest import approx
import numpy as np

from picoquake.data import IMUSample
from picoquake.analisys import *


def test_mean():
    data = [1., 2., 3., 4., 5.]
    assert approx(mean(data), 1e-3) == 3.0

    data = [1., 1., 1., 1., 1.]
    assert approx(mean(data), 1e-3) == 1.0

    data = [0., 0., 0., 0., 0.]
    assert approx(mean(data), 1e-3) == 0.0

    data = [0.]
    assert approx(mean(data), 1e-3) == 0.0

    data = [np.sin(2 * np.pi * i / 120) for i in range(1200)]
    assert approx(mean(data), 1e-3) == 0.0

    data = [5 * np.sin(2 * np.pi * i / 120) for i in range(1200)]
    assert approx(mean(data), 1e-3) == 0.0
    
    data = [5 * np.sin(2 * np.pi * i / 120) + 3.0 for i in range(1200)]
    assert approx(mean(data), 1e-3) == 3.0

    assert mean([]) == 0


def test_detrend():
    data = [1, 2, 3, 4, 5]
    expected = [-2, -1, 0, 1, 2]
    assert detrend(data) == expected

    data = [10, 20, 30, 40, 50]
    expected = [-20, -10, 0, 10, 20]
    assert detrend(data) == expected

    data = [5, 5, 5, 5, 5]
    expected = [0, 0, 0, 0, 0]
    assert detrend(data) == expected

    data = [-1, 0, 1]
    expected = [-1, 0, 1]
    assert detrend(data) == expected

    assert detrend([]) == []


def test_rms():
    data = [1., 2., 3., 4., 5.]
    assert approx(rms(data), 1e-3) == np.sqrt(np.sum(np.square(data)) / 5)
    data_dtr = np.array(data) - np.mean(data)
    assert approx(rms(data, de_trend=True), 1e-3) == np.sqrt(np.sum(np.square(data_dtr)) / 5)

    data = [1, 2, 3]
    assert approx(rms(data), 1e-3) == np.sqrt(np.sum(np.square(data)) / 3)
    data_dtr = np.array(data) - np.mean(data)
    assert approx(rms(data, de_trend=True), 1e-3) == np.sqrt(np.sum(np.square(data_dtr)) / 3)

    data = [1., 1., 1., 1., 1.]
    assert approx(rms(data), 1e-3) == 1.0

    data = [0., 0., 0., 0., 0.]
    assert approx(rms(data), 1e-3) == 0.0

    data = [0.]
    assert approx(rms(data), 1e-3) == 0.0

    data = [np.sin(2 * np.pi * i / 120) for i in range(1200)]
    assert approx(rms(data), 1e-3) == 0.707
    assert approx(rms(data[:120]), 1e-3) == 0.707
    assert approx(rms(data[:60]), 1e-3) == 0.707

    data = [5 * np.sin(2 * np.pi * i / 120) for i in range(1200)]
    assert approx(rms(data), 1e-3) == 5 * 0.707
    assert approx(rms(data[:120]), 1e-3) == 5 * 0.707

    assert rms([]) == 0

    data = ([1., 2., 3., 4., 5.], [1., 2., 3., 4., 5.])
    assert approx(rms(data), 1e-3) == 4.69
    assert approx(rms(data, de_trend=True), 1e-3) == 2.0


def test_imu_rms():
    pass
    samples = [
        IMUSample(0, 1, 2, 5, 8, 1, 4),
        IMUSample(1, 2, 3, 6, 9, 2, 5),
        IMUSample(2, 3, 4, 7, 0, 3, 6),
    ]

    # Test for all axes
    acc_rms, gyro_rms = imu_rms(samples, 'xyz')
    ex_acc_sum = (1 ** 2 + 2 ** 2 + 5 ** 2) + (2 ** 2 + 3 ** 2 + 6 ** 2) + (3 ** 2 + 4 ** 2 + 7 ** 2)
    ex_acc_rms = (ex_acc_sum / 3) ** 0.5
    assert approx(acc_rms, 1e-3) == ex_acc_rms
    ex_gyro_sum = (8 ** 2 + 1 ** 2 + 4 ** 2) + (9 ** 2 + 2 ** 2 + 5 ** 2) + (0 ** 2 + 3 ** 2 + 6 ** 2)
    ex_gyro_rms = (ex_gyro_sum / 3) ** 0.5
    assert approx(gyro_rms, 1e-3) == ex_gyro_rms
    
    acc_rms, gyro_rms = imu_rms(samples, 'xyz', de_trend=True)
    acc_x = np.array([1, 2, 3])
    acc_x = acc_x - np.mean(acc_x)
    acc_y = np.array([2, 3, 4])
    acc_y = acc_y - np.mean(acc_y)
    acc_z = np.array([5, 6, 7])
    acc_z = acc_z - np.mean(acc_z)
    ex_acc_rms = np.sqrt(sum([sum([x[i] ** 2 for x in [acc_x, acc_y, acc_z]]) for i in range(3)]) / 3)
    assert approx(acc_rms, 1e-3) == ex_acc_rms
    gyro_x = np.array([8, 9, 0])
    gyro_x = gyro_x - np.mean(gyro_x)
    gyro_y = np.array([1, 2, 3])
    gyro_y = gyro_y - np.mean(gyro_y)
    gyro_z = np.array([4, 5, 6])
    gyro_z = gyro_z - np.mean(gyro_z)
    ex_gyro_rms = np.sqrt(sum([sum([x[i] ** 2 for x in [gyro_x, gyro_y, gyro_z]]) for i in range(3)]) / 3)
    assert approx(gyro_rms, 1e-3) == ex_gyro_rms

    # Test for x axis only
    acc_rms, gyro_rms = imu_rms(samples, 'x')
    ex_acc_sum = (1 ** 2) + (2 ** 2) + (3 ** 2)
    ex_acc_rms = (ex_acc_sum / 3) ** 0.5
    assert approx(acc_rms, 1e-3) == ex_acc_rms
    ex_gyro_sum = (8 ** 2) + (9 ** 2) + (0 ** 2)
    ex_gyro_rms = (ex_gyro_sum / 3) ** 0.5
    assert approx(gyro_rms, 1e-3) == ex_gyro_rms

    # Test for y axis only
    acc_rms, gyro_rms = imu_rms(samples, 'y')
    ex_acc_sum = (2 ** 2) + (3 ** 2) + (4 ** 2)
    ex_acc_rms = (ex_acc_sum / 3) ** 0.5
    assert approx(acc_rms, 1e-3) == ex_acc_rms
    ex_gyro_sum = (1 ** 2) + (2 ** 2) + (3 ** 2)
    ex_gyro_rms = (ex_gyro_sum / 3) ** 0.5
    assert approx(gyro_rms, 1e-3) == ex_gyro_rms

    # Test for z axis only
    acc_rms, gyro_rms = imu_rms(samples, 'z')
    ex_acc_sum = (5 ** 2) + (6 ** 2) + (7 ** 2)
    ex_acc_rms = (ex_acc_sum / 3) ** 0.5
    assert approx(acc_rms, 1e-3) == ex_acc_rms
    ex_gyro_sum = (4 ** 2) + (5 ** 2) + (6 ** 2)
    ex_gyro_rms = (ex_gyro_sum / 3) ** 0.5
    assert approx(gyro_rms, 1e-3) == ex_gyro_rms

    assert imu_rms([], 'xyz') == (0, 0)

def test_running_rms():
    sine_1 = [np.sin(np.pi * i / 10) for i in range(100)]
    sine_2 = [2 * np.sin(np.pi * i / 10) for i in range(100)]
    sine_3 = [3 * np.sin(np.pi * i / 10) for i in range(100)]
    sine = np.concatenate([sine_1, sine_2, sine_3]).tolist()
    rms_values = running_rms(sine, 4)
    assert approx(rms_values[0], 1e-3) == 0.0
    assert approx(rms_values[20], 1e-3) == 0.707
    assert approx(rms_values[120], 1e-3) == 2 * 0.707
    assert approx(rms_values[220], 1e-3) == 3 * 0.707

    assert running_rms([], 10) == []



"""
Data classes for storing device and acquisition data.
"""

from dataclasses import dataclass, field
from enum import Enum
from hashlib import blake2b
from datetime import datetime
import csv
from typing import Optional
import os

from .configuration import *


class State(Enum):
    IDLE = 0
    SAMPLING = 1
    ERROR = 2


class CommandID(Enum):
    HANDSHAKE = 0
    START_SAMPLING = 1
    STOP_SAMPLING = 2


class PacketID(Enum):
    IMU_DATA = 1
    STATUS = 2
    DEVICE_INFO = 3
    COMMAND = 4


@dataclass
class IMUSample:
    """
    Data class for storing a single IMU sample.

    Attributes:
        count: Sample count.
        acc_x: Accelerometer X value.
        acc_y: Accelerometer Y value.
        acc_z: Accelerometer Z value.
        gyro_x: Gyroscope X value.
        gyro_y: Gyroscope Y value.
        gyro_z: Gyroscope Z value.
    """
    count: int
    acc_x: float
    acc_y: float
    acc_z: float
    gyro_x: float
    gyro_y: float
    gyro_z: float

    def __str__(self) -> str:
        return (f"cnt = {self.count}, "
                f"a_x = {self.acc_x:+.2f}, "
                f"a_y = {self.acc_y:+.2f}, "
                f"a_z = {self.acc_z:+.2f}, "
                f"g_x = {self.gyro_x:+.2f}, "
                f"g_y = {self.gyro_y:+.2f}, "
                f"g_z = {self.gyro_z:+.2f}")


@dataclass
class Status:
    state: State
    temperature: float
    missed_samples: int
    error_code: int

    def __str__(self) -> str:
        return (f"state = {self.state.name}, "
                f"temp = {self.temperature:+.2f}, "
                f"missed = {self.missed_samples}, "
                f"error = {self.error_code}")
    

@dataclass
class DeviceInfo:
    unique_id: str
    firmware: str

    @staticmethod
    def unique_id_to_short_id(unique_id: str) -> str:
        short = blake2b(unique_id.encode(), digest_size=2).hexdigest().upper()
        assert len(short) == 4
        return short
    
    @property
    def short_id(self) -> str:
        return self.unique_id_to_short_id(self.unique_id)

    def __str__(self) -> str:
        return (f"device_id = {self.unique_id.upper()}, "
                f"short_id = {self.short_id}, "
                f"firmware = {self.firmware}")


@dataclass
class AcquisitionData:
    """
    Data class for storing the complete acquisition result,
    including IMU samples, device information, and acquisition configuration.

    Attributes:
        samples: List of IMU samples.
        device: Device information.
        config: Acquisition configuration.
        start_time: Start time of the acquisition.
        skipped_samples: Number of skipped samples due to acquisition issues.
        csv_path: Path to the CSV file containing the data. Used if the data was loaded from a file.
        duration: Duration of the acquisition in seconds.
        num_samples: Number of samples in the acquisition.
        integrity: Whether the acquisition has integrity (no skipped samples).
    
    Methods:
        to_csv: Write the data to a CSV file.
        from_csv: Load the data from a CSV file.
    """
    samples: list[IMUSample]
    device: DeviceInfo
    config: Config
    start_time: datetime
    csv_path: Optional[str] = None

    @property
    def duration(self) -> float:
        return self.num_samples / self.config.sample_rate.param_value
    
    @property
    def num_samples(self) -> int:
        return len(self.samples)
    
    @property
    def skipped_samples(self) -> int:
        return self._check_integrity()
    
    @property
    def integrity(self) -> bool:
        return self.skipped_samples == 0
    
    @property
    def filename(self) -> Optional[str]:
        if self.csv_path is None:
            return None
        else:
            return os.path.basename(self.csv_path)
        
    def re_centre(self, index: int):
        """
        Re-centre the data around a specific index.
        
        Args:
            index: The index to re-centre the data around.
        """
        index = min(max(0, index), len(self.samples) - 1)
        first = self.samples[index].count
        for s in self.samples:
            s.count -= first
    
    def _check_integrity(self) -> int:
        """Counts skipped samples."""
        last_count = self.samples[0].count
        skipped = 0
        for s in self.samples:
            diff = s.count - last_count
            if diff > 1:
                skipped += diff - 1
            last_count = s.count
        return skipped

    def __str__(self) -> str:
        return (f"device = {self.device.short_id}, "
                f"start_time = {self.start_time.isoformat(sep=' ')}, "
                f"num_samples = {self.num_samples}, " 
                f"duration = {self.duration:.2f}s, "
                f"skipped = {self.skipped_samples}")
    
    def to_csv(self, path: str):
        """
        Write the data to a CSV file.

        Args:
            filename: Path to the CSV file.
        """
        metadata = f"# PLab PicoQuake Data\n" \
               f"# Time: {self.start_time.isoformat(sep=' ')}, Device: {self.device.short_id.upper()} ({self.device.unique_id}), " \
               f"FW: {self.device.firmware}\n" \
               f"# Num. samples: {self.num_samples}, Duration: {self.duration} s\n" \
               f"# Config: {self.config}\n" \
               f"# Integrity: {self.integrity}, Skipped samples: {self.skipped_samples}\n"
        with open(path, "w", newline="") as f:
            f.write(metadata)
            writer = csv.writer(f)
            writer.writerow(["count", "a_x", "a_y", "a_z", "g_x", "g_y", "g_z"])
            for sample in self.samples:
                writer.writerow([sample.count, sample.acc_x, sample.acc_y, sample.acc_z,
                                sample.gyro_x, sample.gyro_y, sample.gyro_z])

    @classmethod
    def from_csv(cls, path: str) -> 'AcquisitionData':
        """
        Load the data from a CSV file.

        Args:
            path: Path to the CSV file.

        Returns:
            AcquisitionData: Data loaded from the CSV file.

        Raises:
            ValueError: If an error occurs while parsing the CSV file.
        """
        with open(path, "r") as f:
            reader = csv.reader(f)
            metadata = []
            try:
                for _ in range(5):
                    metadata.append(next(reader))
                start_time = datetime.fromisoformat(metadata[1][0].split(": ")[1])

                unique_id = metadata[1][1].split(": ")[1].split(" ")[1][1:-1]
                try:
                    firmware = metadata[1][2].split(": ")[1]
                except IndexError:
                    firmware = "unknown"
                device = DeviceInfo(unique_id, firmware)

                sample_rate = float(metadata[3][0].split(" = ")[1].split(" ")[0])
                filter = float(metadata[3][1].split(" = ")[1].split(" ")[0])
                acc_range = float(metadata[3][2].split(" = ")[1].split(" ")[0])
                gyro_range = float(metadata[3][3].split(" = ")[1].split(" ")[0])
                config = Config(SampleRate.from_param_value(sample_rate),
                                Filter.from_param_value(filter),
                                AccRange.from_param_value(acc_range),
                                GyroRange.from_param_value(gyro_range))
            except Exception as e:
                raise ValueError(f"Error parsing metadata: {e}")
            next(reader)

            samples = []
            try:
                for row in reader:
                    count, a_x, a_y, a_z, g_x, g_y, g_z = map(float, row)
                    samples.append(IMUSample(int(count), a_x, a_y, a_z, g_x, g_y, g_z))
            except Exception as e:
                raise ValueError(f"Error parsing samples: {e}")
            
            return cls(samples=samples,
                       device=device,
                       config=config,
                       start_time=start_time,
                       csv_path=path)



from typing import cast
import logging
from logging.handlers import RotatingFileHandler
from serial.tools.list_ports import comports
import argparse
import sys
import os
import platform
from datetime import datetime
from time import time, sleep

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

from . import __version__
from .interface import *
from .plot import *


logger = logging.getLogger(__name__)


def equal_with_tolerance(a: float, b: float, tolerance: float = 1e-6) -> bool:
    return abs(a - b) < tolerance


def check_orientation(sample: IMUSample, target: list[float], tol: float = 0.1) -> bool:
    return all(equal_with_tolerance(getattr(sample, axis), val, tol) for axis, val in zip(['acc_x', 'acc_y', 'acc_z'], target))


def _get_log_path(app_name):
    """
    Determines the operating system and returns the appropriate log directory path for the application.

    Args:
    app_name (str): The name of the application.

    Returns:
    str: The full path to the directory where logs should be stored.
    """
    if platform.system() == 'Windows':
        # Windows: Use LOCALAPPDATA
        log_path = os.path.join(os.environ['LOCALAPPDATA'], app_name, 'logs')
    elif platform.system() == 'Linux':
        # Linux: Use a hidden directory in the user's home directory
        log_path = os.path.join(os.path.expanduser('~'), f'.{app_name}', 'logs')
    elif platform.system() == 'Darwin':
        # macOS: Use the Library/Logs directory in the user's home directory
        log_path = os.path.expanduser(f'~/Library/Logs/{app_name}')
    else:
        # Fallback for any other unknown OS
        log_path = os.path.join(os.path.expanduser('~'), f'{app_name}_logs')

    # Create the log directory if it does not exist
    os.makedirs(log_path, exist_ok=True)
    return log_path


def _acquire(args):
    short_id: str = args.short_id
    out: str = args.out
    seconds: float = args.seconds
    sample_rate: float = args.sample_rate
    filter: float = args.filter
    acc_range: float = args.acc_range
    gyro_range: float = args.gyro_range
    autostart: bool = args.autostart
    yes: bool = args.yes

    if not sample_rate >= 2 * filter:
        print("Warning: sample rate should be >= 2 * filter frequency.")
    
    # Check if the output file already exists
    if os.path.isfile(out) and not yes:
        usr = input(f"File {out} already exists. Overwrite? y/n: ")
        if usr.lower() != 'y':
            print("Exiting...")
            sys.exit(0)

    # Find the device
    try:
        device = PicoQuake(short_id)
    except DeviceNotFound:
        print(f"Device with short_id {short_id} not found.")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)

    # Configure device and acquire data
    try:
        device.configure_approx(sample_rate, filter, acc_range, gyro_range)
        config = device.config
        print(f"Configured to: {config}")
    except Exception as e:
        logger.exception(e)
        print(f"Error configuring device: {e}")
        sys.exit(1)
    
    try:
        if not autostart:
            input("\nPress ENTER to start acquisition...\n")
        print("Acquiring...")
        data, exception = device.acquire(seconds)
        print("Done.")
        data.to_csv(out)
        path = os.path.abspath(out)
        print(f"Data written to {path}")  
        if exception is not None:
            raise exception
    except KeyboardInterrupt:
        logger.info("Interrupted by user.")
        print("\nInterrupted by user.")
        sys.exit(1)
    except AcquisitionIncomplete as e:
        print(f"WARNING: Acquisition incomplete: {e}")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)  
    finally:
        device.stop()


def _trigger(args):
    short_id: str = args.short_id
    out: str = args.out
    sample_rate: float = args.sample_rate
    filter: float = args.filter
    acc_range: float = args.acc_range
    gyro_range: float = args.gyro_range
    rms_threshold: float = args.rms_threshold
    pre_seconds: float = args.pre_seconds
    post_seconds: float = args.post_seconds
    source: str = args.source
    axis: str = args.axis
    rms_window: float = args.rms_window
    yes: bool = args.yes

    if not sample_rate >= 2 * filter:
        print("Warning: sample rate should be >= 2 * filter frequency.")
    
    # Check if the output file already exists
    if os.path.isfile(out) and not yes:
        usr = input(f"File {out} already exists. Overwrite? y/n: ")
        if usr.lower() != 'y':
            print("Exiting...")
            sys.exit(0)

    # Find the device
    try:
        device = PicoQuake(short_id)
    except DeviceNotFound:
        print(f"Device with short_id {short_id} not found.")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)

    # Configure device and acquire data
    try:
        device.configure_approx(sample_rate, filter, acc_range, gyro_range)
        config = device.config
        print(f"Configured to: {config}")
        print(f"Trigger set to {rms_threshold} {'g' if source == 'accel' else 'dps'}, "
              f"on {source} axis {axis}, window {rms_window} s.")
    except Exception as e:
        logger.exception(e)
        print(f"Error configuring device: {e}")
        sys.exit(1)

    try:
        print("Waiting for trigger...")
        on_trigger = lambda val: print(f"Triggered at {val:.2f}. Acquiring...")
        data, exception = device.trigger(rms_threshold, pre_seconds, post_seconds, source, axis, rms_window, on_trigger)
        print("Done.")
        data.to_csv(out)
        path = os.path.abspath(out)
        print(f"Data written to {path}")  
        if exception is not None:
            raise exception
    except KeyboardInterrupt:
        logger.info("Interrupted by user.")
        print("\nInterrupted by user.")
        sys.exit(1)
    except AcquisitionIncomplete as e:
        print(f"WARNING: Acquisition incomplete: {e}")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)  
    finally:
        device.stop()


def _run(args):
    with open(args.config, "rb") as f:
        config = tomllib.load(f)

    if "continuous" in config:
        continuous = True
        interval = config["continuous"]["interval"]
    else:
        continuous = False

    while True:
        start_time = time()
        try:
            # output
            if config["output"]["sequential"] and config["output"]["use_timestamp"]:
                print("Error: Configuration cannot have both 'sequential' and 'use_timestamp' set to true.")
                sys.exit(1)
            elif config["output"]["sequential"]:
                path = get_unique_filename(config["output"]["path"])
                overwrite = True
            elif config["output"]["use_timestamp"]:
                path = config["output"]["path"]
                if not os.path.isdir(path):
                    print(f"Error: Directory {path} does not exist.")
                    sys.exit(1)
                path = os.path.join(path, "PQ_" + datetime.now().isoformat() + ".csv")
                overwrite = True
            else:
                path = config["output"]["path"]
                directory, filename = os.path.split(path)
                if directory and not os.path.isdir(directory):
                    print(f"Error: Directory {directory} does not exist.")
                    sys.exit(1)
                overwrite = not config["output"]["confirm_overwrite"]

            # run command
            if "acquire" in config and not "trigger" in config:
                print("Run -> Acquire")
                pass_args = {**config["device"], **config["config"], **config["acquire"],
                            "out": path, "yes": overwrite, "autostart": True}
                _acquire(argparse.Namespace(**pass_args))
            elif "trigger" in config and not "acquire" in config:
                print("Run -> Trigger")
                pass_args = {**config["device"], **config["config"], **config["trigger"],
                            "out": path, "yes": overwrite}
                _trigger(argparse.Namespace(**pass_args))
            else:
                print("Error: Configuration must contain either 'acquire' or 'trigger' section.")
                sys.exit(1)

            # wait
            if continuous:
                elapsed = time() - start_time
                if elapsed < interval:
                    sleep(interval - elapsed)
            else:
                break

        except KeyboardInterrupt:
            print("\nInterrupted by user.")
            sys.exit(0)
        except (AcquisitionIncomplete, AcquisitionDataCorrupted) as e:
            logger.exception(e)
            print(f"Error: {e}")
        except Exception as e:
            logger.exception(e)
            print(f"Error: {e}")
            sys.exit(1)


def _live_display(args):
    short_id: str = args.short_id
    interval: float = args.interval
    if interval < 0.1:
        interval = 0.1
    elif interval > 10:
        interval = 10
    
    try:
        device = PicoQuake(short_id)
    except DeviceNotFound:
        print(f"Device with short_id {short_id} not found.")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)
    try:   
        device.configure(SampleRate.hz_12_5, Filter.hz_42, AccRange.g_4, GyroRange.dps_250)
        device.start_continuos()
        while True:
            sample = device.read_last()
            print(sample)
            sleep(interval)
    except KeyboardInterrupt:
        logger.info("Interrupted by user.")
        print("\nInterrupted by user.")
        sys.exit(0)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)
    finally:
        device.stop()


def _plot_psd(args):
    csv_path: str = args.csv_path
    output: str = args.output
    axis: str = args.axis
    freq_min: float = args.fmin
    freq_max: float = args.fmax
    peaks: bool = args.peaks
    title: str = args.title

    output = output if output != '.' else os.path.splitext(csv_path)[0] + "_psd.png"

    try:
        result = AcquisitionData.from_csv(csv_path)
    except Exception as e:
        logger.exception(e)
        print(f"Error loading file: {e}")
        sys.exit(1)
    try:
        plot_psd(result, output, axis, freq_min, freq_max, peaks, title)
        print(f"Plot saved to {output}")
    except ModuleNotFoundError:
        print("Plotting not supported. To enable install with 'pip install picoquake[plot]'.")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)

def _plot_fft(args):
    csv_path: str = args.csv_path
    output: str = args.output
    axis: str = args.axis
    freq_min: float = args.fmin
    freq_max: float = args.fmax
    peaks: bool = args.peaks
    title: str = args.title

    output = output if output != '.' else os.path.splitext(csv_path)[0] + "_fft.png"

    try:
        result = AcquisitionData.from_csv(csv_path)
    except Exception as e:
        logger.exception(e)
        print(f"Error loading file: {e}")
        sys.exit(1)
    try:
        plot_fft(result, output, axis, freq_min, freq_max, peaks, title)
        print(f"Plot saved to {output}")
    except ModuleNotFoundError:
        print("Plotting not supported. To enable install with 'pip install picoquake[plot]'.")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)

def _plot(args):
    csv_path: str = args.csv_path
    output: str = args.output
    axis: str = args.axis
    time_start: float = args.tstart
    time_end: float = args.tend
    title: str = args.title
    rms: bool = args.rms
    rms_win: float = args.rms_win
    rms_detrend: bool = args.rms_detrend

    output = output if output != '.' else os.path.splitext(csv_path)[0] + "_plot.png"

    try:
        result = AcquisitionData.from_csv(csv_path)
    except Exception as e:
        logger.exception(e)
        print(f"Error loading file: {e}")
        sys.exit(1)
    try:
        plot(result, output, axis, time_start, time_end, title, rms, rms_win, rms_detrend)
        print(f"Plot saved to {output}")
    except ModuleNotFoundError:
        print("Plotting not supported. To enable install with 'pip install picoquake[plot]'.")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)



def _list_devices(args):
    all_ports = args.all
    ports = comports()
    devices = []
    for p in ports:
        if p.vid == VID and p.pid == PID and p.serial_number:
            short_id = DeviceInfo.unique_id_to_short_id(p.serial_number)
            devices.append((short_id, p))
            print(f"PicoQuake {short_id}: {p.device}")
        elif all_ports:
            print(f"Unknown device: {p.device}, description: {p.description}")
    if not devices:
        print("No PicoQuake devices found.")


def _info(args):
    short_id: str = args.short_id
    try:
        device = PicoQuake(short_id)
    except DeviceNotFound:
        print(f"Device with short_id {short_id} not found.")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)
    try:
        info = device.device_info
        print(info)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)
    finally:
        device.stop()


def _test(args):
    tol = 0.1
    short_id: str = args.short_id
    try:
        device = PicoQuake(short_id)
    except DeviceNotFound:
        print(f"Device with short_id {short_id} not found.")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)
    try:
        device.configure(SampleRate.hz_12_5, Filter.hz_42, AccRange.g_4, GyroRange.dps_250)
        device.start_continuos()
        print("Point Z up...", end="", flush=True)
        while True:
            sample = cast(IMUSample, device.read_last())
            if check_orientation(sample, [0, 0, 1], tol):
                print("OK")
                break
            sleep(0.1)
        print("Point X up...", end="", flush=True)
        while True:
            sample = cast(IMUSample, device.read_last())
            if check_orientation(sample, [1, 0, 0], tol):
                print("OK")
                break
            sleep(0.1)
        print("Point Y down...", end="", flush=True)
        while True:
            sample = cast(IMUSample, device.read_last())
            if check_orientation(sample, [0, -1, 0], tol):
                print("OK")
                break
            sleep(0.1)
        print("Successful.")
    except KeyboardInterrupt:
        logger.info("Interrupted by user.")
        print("\nInterrupted by user.")
        sys.exit(0)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)
    finally:
        device.stop()


def _reboot_to_bootsel(args):
    short_id: str = args.short_id
    try:
        device = PicoQuake(short_id)
    except DeviceNotFound:
        print(f"Device with short_id {short_id} not found.")
        sys.exit(1)
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)
    try:
        device.reboot_to_bootsel()
        print("Device rebooted to BOOTSEL.")
    except Exception as e:
        logger.exception(e)
        print(f"Error: {e}")
        sys.exit(1)


def main():
    main_parser = argparse.ArgumentParser(description="PicoQuake CLI.")
    main_parser.add_argument("-v", "--version", action="version", version=f"picoquake {__version__}")
    main_parser.add_argument("--verbose", action="store_true", help="Print log messages to console.")
    main_parser.add_argument("--debug", action="store_true", help="Set log level to debug.")
    subparsers = main_parser.add_subparsers(required=True, dest="command")

    # acquire
    acquire_parser = subparsers.add_parser("acquire", help="Acquire data from a PicoQuake device.",
                                           fromfile_prefix_chars='@')
    acquire_parser.add_argument("short_id", help="The 4 character ID of the device. Found on the label.")
    acquire_parser.add_argument("out", help="The output CSV file.")
    acquire_parser.add_argument("-s", "--seconds", type=float, default=2.0,
                                help="Duration of the acquisition in seconds.")
    acquire_parser.add_argument("-r", "--sample_rate", type=float, default=200.0,
                                help="Sample rate in Hz. Range 12.5 - 4000 Hz. Closest available selected.")
    acquire_parser.add_argument("-f", "--filter", type=float, default=42.0,
                                help="Filter frequency in Hz. Range 42 - 3979 Hz. Closest available selected.")
    acquire_parser.add_argument("-ar", "--acc_range", type=float, default=4.0,
                                help="Acceleration range in g. Range 2 - 16 g. Closest available selected.")
    acquire_parser.add_argument("-gr", "--gyro_range", type=float, default=1000.0,
                                help="Gyro range in dps. Range 15.625 - 2000 dps. Closest available selected.")
    acquire_parser.add_argument("-a", "--autostart", action="store_true",
                                help="Start acquisition without user confirmation.")
    acquire_parser.add_argument("-y", "--yes", action="store_true",
                                help="Skip overwrite prompt.")
    acquire_parser.set_defaults(func=_acquire)

    # trigger
    trigger_parser = subparsers.add_parser("trigger", help="Trigger acquisition based on RMS threshold.",
                                           fromfile_prefix_chars='@')
    trigger_parser.add_argument("short_id", help="The 4 character ID of the device. Found on the label.")
    trigger_parser.add_argument("out", help="The output CSV file.")
    trigger_parser.add_argument("-r", "--sample_rate", type=float, default=200.0,
                                help="Sample rate in Hz. Range 12.5 - 4000 Hz. Closest available selected.")
    trigger_parser.add_argument("-f", "--filter", type=float, default=42.0,
                                help="Filter frequency in Hz. Range 42 - 3979 Hz. Closest available selected.")
    trigger_parser.add_argument("-ar", "--acc_range", type=float, default=4.0,
                                help="Acceleration range in g. Range 2 - 16 g. Closest available selected.")
    trigger_parser.add_argument("-gr", "--gyro_range", type=float, default=1000.0,
                                help="Gyro range in dps. Range 15.625 - 2000 dps. Closest available selected.")
    trigger_parser.add_argument("--rms_threshold", type=float, help="RMS threshold for triggering.")
    trigger_parser.add_argument("--pre_seconds", type=float, default=0.0,
                                help="Duration before trigger in seconds.")
    trigger_parser.add_argument("--post_seconds", type=float, default=1.0,
                                help="Duration after trigger in seconds.")
    trigger_parser.add_argument("--source", choices=["accel", "gyro"], default="accel",
                                help="Source for triggering.")
    trigger_parser.add_argument("-a", "--axis", default="xyz", help="Trigger axis, must be 'x', 'y', 'z', or a combination")
    trigger_parser.add_argument("--rms_window", type=float, default=1.,
                                help="Window size for RMS calculation.")
    trigger_parser.add_argument("-y", "--yes", action="store_true",
                                help="Skip overwrite prompt.")
    trigger_parser.set_defaults(func=_trigger)

    # run
    run_parser = subparsers.add_parser("run", help="Run acquisition from config file.")
    run_parser.add_argument("config", help="The configuration TOML file.")
    run_parser.set_defaults(func=_run)

    # display
    live_disp_parser = subparsers.add_parser("display", help="Display live data from device.")
    live_disp_parser.add_argument("short_id", help="The 4 character ID of the device. Found on the label.")
    live_disp_parser.add_argument("-i", "--interval", type=float, default=1.0,
                                  help="Interval between samples in seconds. Range 0.1 - 10 s.")
    live_disp_parser.set_defaults(func=_live_display)

    # list devices
    list_devices_parser = subparsers.add_parser("list", help="List connected PicoQuake devices.")
    list_devices_parser.add_argument("-a", "--all", action="store_true", help="List all serial ports.")
    list_devices_parser.set_defaults(func=_list_devices)

    # device info
    info_parser = subparsers.add_parser("info", help="Display device information.")
    info_parser.add_argument("short_id", help="The 4 character ID of the device. Found on the label.")
    info_parser.set_defaults(func=_info)

    # test
    test_parser = subparsers.add_parser("test", help="Test device.")
    test_parser.add_argument("short_id", help="The 4 character ID of the device. Found on the label.")
    test_parser.set_defaults(func=_test)

    # reboot to bootsel
    reboot_bootsel = subparsers.add_parser("bootsel", help="Reboot device to BOOTSEL.")
    reboot_bootsel.add_argument("short_id", help="The 4 character ID of the device. Found on the label.")
    reboot_bootsel.set_defaults(func=_reboot_to_bootsel)

    # plot PSD
    fftplot_parser = subparsers.add_parser("plot_psd", help="Plot Power Spectral Density of acquired data.")
    fftplot_parser.add_argument("csv_path", help="The CSV file containing the acquired data.")
    fftplot_parser.add_argument("output", help="The output file to save the plot to. '.' to save next to the data file.")
    fftplot_parser.add_argument("-a", "--axis", default="xyz", help="Axis to plot, must be 'x', 'y', 'z', or a combination")
    fftplot_parser.add_argument("--fmin", type=float, default=0.0, help="Minimum frequency to plot.")
    fftplot_parser.add_argument("--fmax", type=float, default=1000.0, help="Maximum frequency to plot.")
    fftplot_parser.add_argument("--peaks", action="store_true", help="Annotate peaks on the plot.")
    fftplot_parser.add_argument("--title", help="Title of the plot.", default=None)
    fftplot_parser.set_defaults(func=_plot_psd)

    # plot FFT
    fftplot_parser = subparsers.add_parser("plot_fft", help="Plot Fast Fourier Transform of acquired data.")
    fftplot_parser.add_argument("csv_path", help="The CSV file containing the acquired data.")
    fftplot_parser.add_argument("output", help="The output file to save the plot to. '.' to save next to the data file.")
    fftplot_parser.add_argument("-a", "--axis", default="xyz", help="Axis to plot, must be 'x', 'y', 'z', or a combination")
    fftplot_parser.add_argument("--fmin", type=float, default=0.0, help="Minimum frequency to plot.")
    fftplot_parser.add_argument("--fmax", type=float, default=1000.0, help="Maximum frequency to plot.")
    fftplot_parser.add_argument("--peaks", action="store_true", help="Annotate peaks on the plot.")
    fftplot_parser.add_argument("--title", help="Title of the plot.", default=None)
    fftplot_parser.set_defaults(func=_plot_fft)

    # plot
    plot_parser = subparsers.add_parser("plot", help="Plot acquired data (time series).")
    plot_parser.add_argument("csv_path", help="The CSV file containing the acquired data.")
    plot_parser.add_argument("output", help="The output file to save the plot to. '.' to save next to the data file.")
    plot_parser.add_argument("-a", "--axis", default="xyz", help="Axis to plot, must be 'x', 'y', 'z', or a combination")
    plot_parser.add_argument("--tstart", type=float, default=float("-inf"), help="Start time of the plot.")
    plot_parser.add_argument("--tend", type=float, default=float("inf"), help="End time of the plot.")
    plot_parser.add_argument("--title", help="Title of the plot.", default=None)
    plot_parser.add_argument("--rms", action="store_true", help="Plot RMS values.")
    plot_parser.add_argument("--rms_win", type=float, default=1.0, help="Window size for RMS calculation.")
    plot_parser.add_argument("--rms_detrend", action="store_true", help="Detrend data before RMS calculation.")
    plot_parser.set_defaults(func=_plot)

    args = main_parser.parse_args()

    # logging
    log_path = os.path.join(_get_log_path("picoquake"), "picoquake.log")
    file_handler =RotatingFileHandler(log_path, mode='a', maxBytes=15*1024*1024, backupCount=5)
    file_handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
    logging.getLogger().addHandler(file_handler)
    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)
    if args.verbose:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter('%(levelname)s - %(message)s'))
        logging.getLogger().addHandler(stream_handler)
    
    args.func(args)
    
if __name__ == '__main__':
    main()



"""
This example demonstrates how to trigger an acquisition based on a root mean square (RMS) threshold.
Sample rate is handled by the device and is very accurate.
"""

import logging

import picoquake

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    # Create a PicoQuake device
    device = picoquake.PicoQuake("c6e3")

    # Configure acquisition
    device.configure(
        sample_rate=picoquake.SampleRate.hz_1000,
        filter_hz=picoquake.Filter.hz_213,
        acc_range=picoquake.AccRange.g_16,
        gyro_range=picoquake.GyroRange.dps_2000,
    )

    # Acquire data from the device
    data, exception = device.trigger(rms_threshold=0.5,
                                     pre_seconds=1,
                                     post_seconds=3,
                                     source="accel",
                                     axis="xyz",
                                     rms_window=0.1)

    # Stop the device
    device.stop()

    # Check if an exception occurred
    if exception is not None:
        print(f"Exception: {exception}")

    # Print the acquired data
    print(f"Data: {data}")

    # Save to a CSV file
    data.to_csv("acquisition_triggered.csv")



"""
Exceptions used by the library.
"""

class HandshakeError(Exception):
    """
    Raised when handshake with the device fails.
    """
    pass


class ConnectionError(Exception):
    """
    Raised when connection to the device fails.
    """
    pass


class DeviceNotFound(Exception):
    """
    Raised when device with specified short ID is not found.
    """
    pass


class DeviceError(Exception):
    """
    Raised when connected device returns an error.

    Attributes:
        error_code: The error code returned by the device.
    """
    
    def __init__(self, error_code: int):
        self.error_code: int = error_code
        super().__init__(f"Device error: {error_code}")

class AcquisitionIncomplete(Exception):
    """
    Raised when acquisition is incomplete.
    """
    pass

class AcquisitionDataCorrupted(Exception):
    """
    Raised when acquisition data is corrupted.
    """
    pass



from itertools import permutations

from .data import *
from .utils import get_axis_combinations
from .analisys import running_rms

def plot_psd(result: AcquisitionData, output_file: str, axis: str = "xyz",
             freq_min: float = 0, freq_max: Optional[float] = None,
             show_peaks: bool = False, title=None) -> None:
    import numpy as np
    import matplotlib.pyplot as plt
    from scipy.signal import welch, find_peaks

    combinations = get_axis_combinations("xyz")
    if axis not in combinations:
        raise ValueError("Invalid axis, must be 'x', 'y', 'z', or a combination.")
    if not result.integrity:
        print(f"Warning: Data integrity compromised, {result.skipped_samples} samples skipped.")

    # check Nyquist criterion
    if result.config.sample_rate.param_value < 2 * result.config.filter.param_value:
        print(f"Warning: sample rate {result.config.sample_rate.param_value} Hz is "
              f"not >= 2 * filter frequency {result.config.filter.param_value} Hz.")
    
    # check frequency range
    if freq_max is None:
        freq_max = result.config.sample_rate.param_value // 2
    elif freq_max > result.config.sample_rate.param_value // 2:
        print(f"Warning: freq_max ({freq_max} Hz) is greater "
              f"than 0.5 x sample rate ({result.config.sample_rate.param_value} Hz). "
              f"Limiting to {result.config.sample_rate.param_value // 2} Hz.")
        freq_max = result.config.sample_rate.param_value // 2

    if freq_min >= freq_max:
        raise ValueError("freq_min must be less than freq_max.")

    acc_x = np.array([s.acc_x for s in result.samples])
    acc_y = np.array([s.acc_y for s in result.samples])
    acc_z = np.array([s.acc_z for s in result.samples])

    # calculate segment length based on plot frequency range
    nperseg = min(int((100 * result.config.sample_rate.param_value) // (freq_max - freq_min)), len(acc_x))

    plt.figure(figsize=(10, 8))  # Increase figure size. You can adjust the values as needed.
    for ax, acc, color in zip(['x', 'y', 'z'], [acc_x, acc_y, acc_z], ["red", "green", "blue"]):
        if ax in axis:
            f, p_den = welch(acc, fs=result.config.sample_rate.param_value, nperseg=nperseg, scaling="density")
            mask = (f >= freq_min) & (f <= freq_max)
            f = f[mask]
            p_den = p_den[mask]
            plt.plot(f, p_den, label=ax, linewidth=1.0, color=color)
            if show_peaks:
                peaks, _ = find_peaks(p_den)
                if len(peaks) > 0:
                    peak_heights = p_den[peaks]
                    top_peaks = peaks[np.argsort(peak_heights)[-3:]]
                    for peak in top_peaks:
                        plt.annotate(f'{f[peak]:.1f} Hz', 
                                    (f[peak], p_den[peak]), 
                                    textcoords="offset points", 
                                    xytext=(0,30), 
                                    ha='center', 
                                    color=color,
                                    arrowprops=dict(facecolor=color, shrink=0.1, width=3, headwidth=6, headlength=6),
                                    bbox=dict(boxstyle="round,pad=0.3", edgecolor=color, facecolor='white', alpha=1.0))

    if title is not None:
        plt.title(title, pad=40)
    elif result.filename is not None:
        plt.title(result.filename, pad=40)
    plt.xlabel("Frequency [Hz]")
    plt.ylabel("PSD [g$^2$/Hz]")
    plt.yscale("log")
    plt.grid(True, "both")
    plt.minorticks_on()
    plt.legend(loc="upper right")
    plt.autoscale(enable=True, axis='y', tight=False)
    plt.savefig(output_file, dpi=200)

def plot_fft(result: AcquisitionData, output_file: str, axis: str = "xyz",
             freq_min: float = 0, freq_max: Optional[float] = None,
             show_peaks: bool = False, title=None) -> None:
    import numpy as np
    import matplotlib.pyplot as plt
    from scipy.fft import fft, fftfreq
    from scipy.signal import find_peaks
    from scipy.signal.windows import hann

    combinations = get_axis_combinations("xyz")
    if axis not in combinations:
        raise ValueError("Invalid axis, must be 'x', 'y', 'z', or a combination.")
    if not result.integrity:
        print(f"Warning: Data integrity compromised, {result.skipped_samples} samples skipped.")

    # check Nyquist criterion
    if result.config.sample_rate.param_value < 2 * result.config.filter.param_value:
        print(f"Warning: sample rate {result.config.sample_rate.param_value} Hz is "
              f"not >= 2 * filter frequency {result.config.filter.param_value} Hz.")
    
    # check frequency range
    if freq_max is None:
        freq_max = result.config.sample_rate.param_value // 2
    elif freq_max > result.config.sample_rate.param_value // 2:
        print(f"Warning: freq_max ({freq_max} Hz) is greater "
              f"than 0.5 x sample rate ({result.config.sample_rate.param_value} Hz). "
              f"Limiting to {result.config.sample_rate.param_value // 2} Hz.")
        freq_max = result.config.sample_rate.param_value // 2

    if freq_min >= freq_max:
        raise ValueError("freq_min must be less than freq_max.")

    acc_x = np.array([s.acc_x for s in result.samples])
    acc_y = np.array([s.acc_y for s in result.samples])
    acc_z = np.array([s.acc_z for s in result.samples])


    plt.figure(figsize=(10, 8))  # Increase figure size. You can adjust the values as needed.
    for ax, acc, color in zip(['x', 'y', 'z'], [acc_x, acc_y, acc_z], ["red", "green", "blue"]):
        if ax in axis:
            N = len(acc)
            T = 1.0 / result.config.sample_rate.param_value
            # Apply a window function
            window = hann(len(acc))
            acc = acc * window
            yf = fft(acc)
            xf = fftfreq(N, T)
            mask = (xf >= freq_min) & (xf <= freq_max)
            xf = xf[mask]
            yf = yf[mask]
            plt.plot(xf, 2.0/N * np.abs(yf[0:N//2]), label=ax, linewidth=1.0, color=color)
            if show_peaks:
                peaks, _ = find_peaks(2.0/N * np.abs(yf[0:N//2]))
                if len(peaks) > 0:
                    peak_heights = 2.0/N * np.abs(yf[0:N//2])[peaks]
                    top_peaks = peaks[np.argsort(peak_heights)[-3:]]
                    for peak in top_peaks:
                        plt.annotate(f'{xf[peak]:.1f} Hz', 
                                    (xf[peak], 2.0/N * np.abs(yf[0:N//2])[peak]), 
                                    textcoords="offset points", 
                                    xytext=(0,30), 
                                    ha='center', 
                                    color=color,
                                    arrowprops=dict(facecolor=color, shrink=0.1, width=3, headwidth=6, headlength=6),
                                    bbox=dict(boxstyle="round,pad=0.3", edgecolor=color, facecolor='white', alpha=1.0))


    if title is not None:
        plt.title(title, pad=40)
    elif result.filename is not None:
        plt.title(result.filename, pad=40)
    plt.xlabel("Frequency [Hz]")
    plt.ylabel("Amplitude [g]")
    plt.yscale("log")
    plt.grid(True, "both")
    plt.minorticks_on()
    plt.legend(loc="upper right")
    plt.autoscale(enable=True, axis='y', tight=False)
    plt.savefig(output_file, dpi=200)

def plot(result: AcquisitionData, output_file: str, axis: str = "xyz",
         tstart: float = float("-inf"), tend: float = float("inf"), title=None,
         rms: bool=False, rms_win: float=1.0, rms_detrend: bool=False) -> None:
    import numpy as np
    import matplotlib.pyplot as plt

    combinations = get_axis_combinations("xyz")
    if axis not in combinations:
        raise ValueError("Invalid axis, must be 'x', 'y', 'z', or a combination.")
    if not result.integrity:
        print(f"Warning: Data integrity compromised, {result.skipped_samples} samples skipped.")

    acc_x = np.array([s.acc_x for s in result.samples])
    acc_y = np.array([s.acc_y for s in result.samples])
    acc_z = np.array([s.acc_z for s in result.samples])

    plt.figure(figsize=(10, 6))  # Increase figure size. You can adjust the values as needed.

    dt = 1 / result.config.sample_rate.param_value
    t = np.array([dt * s.count for s in result.samples])
    if(tend == None):
        tend = len(acc_x) / result.config.sample_rate.param_value
    mask = (t >= tstart) & (t <= tend)
    t = t[mask]
    for ax, acc, color, rms_color in zip(['x', 'y', 'z'], [acc_x, acc_y, acc_z], ["red", "green", "blue"], ["orange", "lightgreen", "lightblue"]):
        if ax in axis:
            acc = acc[mask]
            plt.plot(t, acc, label=ax, linewidth=1.0, color=color)
            if rms:
                rms_data = running_rms(acc.tolist(), int(rms_win * result.config.sample_rate.param_value),
                                       de_trend=rms_detrend)
                plt.plot(t, rms_data, label=f"{ax} RMS", linewidth=2.0, color=rms_color, linestyle="--")

    if title is not None:
        plt.title(title, pad=20)
    elif result.filename is not None:
        plt.title(result.filename, pad=20)
    plt.xlabel("Time [s]")
    plt.ylabel("Acceleration [g]")
    plt.grid(True, "both")
    plt.minorticks_on()
    plt.legend(loc="upper right")
    plt.autoscale(enable=True, axis='y', tight=False)
    plt.savefig(output_file, dpi=200)





from enum import Enum
from dataclasses import dataclass
from typing import TypeVar, Type, Union

T = TypeVar('T', bound='ConfigEnum')

class ConfigEnum(Enum):
    """
    Base class for configuration enums.
    Each member has two elements: an index and a parameter value.
    Parameter value can be either an integer or a float.

    Attributes:
        index: Index of the member.
        param_value: Parameter value of the member.

    Methods:
        from_index: Return member with specified index.
        from_param_value: Return member with specified parameter value.
        find_closest: Return member with parameter value closest to specified value.
    """

    def __new__(cls, *args):
        if not len(args) == 2:
            raise ValueError(f"Member must have 2 elements: {args}")
        if not isinstance(args[0], int):
            raise ValueError(f"First element must be an integer: {args}")
        if not isinstance(args[1], (int, float)):
            raise ValueError(f"Second element must be an integer or float {args}")
        obj = object.__new__(cls)
        obj._value_ = args
        return obj

    @classmethod
    def from_index(cls: Type[T], index: int) -> T:
        """Return member with specified index."""
        for member in cls:
            if member.value[0] == index:
                return member
        raise ValueError(f"Invalid index: {index}")

    @classmethod
    def from_param_value(cls: Type[T], value: Union[float, int], tolerance: float = 1e-5) -> T:
        """Return member with specified parameter value."""
        for member in cls:
            if isinstance(value, float):
                if abs(member.value[1] - value) < tolerance:
                    return member
            elif isinstance(value, int):
                if member.value[1] == value:
                    return member
        raise ValueError(f"Invalid value: {value}")
    
    @classmethod
    def find_closest(cls: Type[T], value: Union[float, int]) -> T:
        """Return member with parameter value closest to specified value."""
        if not list(cls):
                raise ValueError("No members in the enum")
        closest = next(iter(cls))
        closest_diff = float('inf')
        for member in cls:
            diff = abs(member.value[1] - value)
            if diff < closest_diff:
                closest = member
                closest_diff = diff
        return closest
    
    @property
    def index(self) -> int:
        return self.value[0]
    
    @property
    def param_value(self) ->  Union[float, int]:
        return self.value[1]


class SampleRate(ConfigEnum):
    hz_12_5 = 0, 12.5
    hz_25 = 1, 25.0
    hz_50 = 2, 50.0
    hz_100 = 3, 100.0
    hz_200 = 4, 200.0
    hz_500 = 5, 500.0
    hz_1000 = 6, 1000.0
    hz_2000 = 7, 2000.0
    hz_4000 = 8, 4000.0


class Filter(ConfigEnum):
    hz_42 = 0, 42
    hz_84 = 1, 84
    hz_126 = 2, 126
    hz_170 = 3, 170
    hz_213 = 4, 213
    hz_258 = 5, 258
    hz_303 = 6, 303
    hz_348 = 7, 348
    hz_394 = 8, 394
    hz_441 = 9, 441
    hz_488 = 10, 488
    hz_536 = 11, 536
    hz_585 = 12, 585
    hz_634 = 13, 634
    hz_684 = 14, 684
    hz_734 = 15, 734
    hz_785 = 16, 785
    hz_837 = 17, 837
    hz_890 = 18, 890
    hz_943 = 19, 943
    hz_997 = 20, 997
    hz_1051 = 21, 1051
    hz_1107 = 22, 1107
    hz_1163 = 23, 1163
    hz_1220 = 24, 1220
    hz_1277 = 25, 1277
    hz_1336 = 26, 1336
    hz_1395 = 27, 1395
    hz_1454 = 28, 1454
    hz_1515 = 29, 1515
    hz_1577 = 30, 1577
    hz_1639 = 31, 1639
    hz_1702 = 32, 1702
    hz_1766 = 33, 1766
    hz_1830 = 34, 1830
    hz_1896 = 35, 1896
    hz_1962 = 36, 1962
    hz_2029 = 37, 2029
    hz_2097 = 38, 2097
    hz_2166 = 39, 2166
    hz_2235 = 40, 2235
    hz_2306 = 41, 2306
    hz_2377 = 42, 2377
    hz_2449 = 43, 2449
    hz_2522 = 44, 2522
    hz_2596 = 45, 2596
    hz_2671 = 46, 2671
    hz_2746 = 47, 2746
    hz_2823 = 48, 2823
    hz_2900 = 49, 2900
    hz_2978 = 50, 2978
    hz_3057 = 51, 3057
    hz_3137 = 52, 3137
    hz_3217 = 53, 3217
    hz_3299 = 54, 3299
    hz_3381 = 55, 3381
    hz_3464 = 56, 3464
    hz_3548 = 57, 3548
    hz_3633 = 58, 3633
    hz_3718 = 59, 3718
    hz_3805 = 60, 3805
    hz_3892 = 61, 3892
    hz_3979 = 62, 3979


class AccRange(ConfigEnum):
    g_2 = 0, 2
    g_4 = 1, 4
    g_8 = 2, 8
    g_16 = 3, 16


class GyroRange(ConfigEnum):
    dps_15_625 = 0, 15.625
    dps_31_25 = 1, 31.25
    dps_62_5 = 2, 62.5
    dps_125 = 3, 125.0
    dps_250 = 4, 250.0
    dps_500 = 5, 500.0
    dps_1000 = 6, 1000.0
    dps_2000 = 7, 2000.0


@dataclass
class Config:
    sample_rate: SampleRate
    filter: Filter
    acc_range: AccRange
    gyro_range: GyroRange

    def __str__(self):
        return f"sample_rate = {self.sample_rate.param_value} Hz, " \
               f"filter = {self.filter.param_value} Hz, " \
               f"acc_range = {self.acc_range.param_value} g, " \
               f"gyro_range = {self.gyro_range.param_value} dps"



"""
This example demonstrates how to continuously read data from device.
Sample rate is handled by this script and may not be accurate due to system load.
Use this method to acquire data in real-time.
"""

from time import sleep
import logging

import picoquake

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    # Create a PicoQuake device
    device = picoquake.PicoQuake("d4e9")

    # Configure acquisition
    device.configure(
        sample_rate=picoquake.SampleRate.hz_100,
        filter_hz=picoquake.Filter.hz_42,
        acc_range=picoquake.AccRange.g_16,
        gyro_range=picoquake.GyroRange.dps_2000,
    )

    # Alternatively, configure from float values.
    # The closest available values will be used.
    device.configure_approx(
        sample_rate=1000,
        filter_hz=200,
        acc_range=16,
        gyro_range=2000,
    )

    # Start continuous acquisition
    device.start_continuos()

    # Read from device
    try:
        while True:
            sample = device.read_last()
            print(f"Sample: {sample}")
            sleep(1)
    except KeyboardInterrupt:
        print("Stopped by user.")
    finally:
        device.stop()
    


from typing import List
from random import uniform
import tempfile
from pytest import approx

from picoquake.data import *
from picoquake.configuration import *


def test_device_info():
    device_info = DeviceInfo(
        unique_id="E66368254F89A225",
        firmware="1.0.0")
    
    assert device_info.short_id == "C6E3"


def test_acquisition_data():
    device_info = DeviceInfo(
        unique_id="E66368254F89A225",
        firmware="1.0.0")
    
    config = Config(
        sample_rate=SampleRate.hz_1000,
        filter=Filter.hz_394,
        acc_range=AccRange.g_16,
        gyro_range=GyroRange.dps_2000,
    )

    samples: List[IMUSample] = []
    N = 1000
    for i in range(-N, N):
        samples.append(IMUSample(
            count=i,
            acc_x=uniform(-1, 1),
            acc_y=uniform(-2, 2),
            acc_z=uniform(-3, 3),
            gyro_x=uniform(-100, 100),
            gyro_y=uniform(-200, 200),
            gyro_z=uniform(-300, 300),
        ))

    data = AcquisitionData(
        samples=samples,
        device=device_info,
        config=config,
        start_time=datetime.now())
    
    assert data.num_samples == 2 * N
    assert data.integrity == True
    assert data.duration == approx(2 * N / 1000, 1e-3)
    
    with tempfile.NamedTemporaryFile(mode='w') as f:
        path = f.name
        print(path)
        data.csv_path = path
        data.to_csv(path)
        loaded_data = AcquisitionData.from_csv(path)
        assert data == loaded_data

    assert data.samples[0].count == -N
    data.re_centre(0)
    assert data.samples[0].count == 0
    data.re_centre(100)
    assert data.samples[0].count == -100

    data.samples.pop(N)
    assert data.integrity == False



"""
This module implements various data analysis functions.
"""

from typing import List, Tuple, Union

from .data import IMUSample


def mean(data: Union[List[float], List[int]]) -> float:
    """
    Calculate the mean of a list of values.
    
    Args:
    data: List of values.

    Returns:
    The mean of the values.
    """

    if len(data) == 0:
        return 0
    return sum(data) / len(data)


def detrend(data: Union[List[float], List[int]]) -> Union[List[float], List[int]]:
    """
    Remove the trend from a list of values.
    
    Args:
    data: List of values.

    Returns:
    List with the trend removed.
    """

    m = mean(data)
    return [x - m for x in data]


def rms(data: Union[Union[List[float], List[int]], Tuple[Union[List[float], List[int]], ...]],
        de_trend: bool=False) -> float:
    """
    Calculate the root mean square of a list of values.
    
    Args:
    data: List of values or a tuple of lists of values. 
          If a tuple is provided, the squares of the values in each list are summed.
    de_trend: If True, remove the trend from the data.

    Returns:
    The root mean square of the values.
    """
    if len(data) == 0:
        return 0
    if isinstance(data, tuple):
        new_data = [x.copy() for x in data]
    elif isinstance(data, list):
        new_data = [data.copy()]
    else:
        raise ValueError("Invalid data type. Must be a list or a tuple of lists.")
    if len(set([len(x) for x in new_data])) != 1:
        raise ValueError("All lists must have the same length.")
    length = len(new_data[0])
    if de_trend:
        new_data = [detrend(x) for x in new_data]
    return (sum([sum([x[i] ** 2 for x in new_data]) for i in range(length)]) / length) ** 0.5


def imu_rms(samples: List[IMUSample], axes: str, de_trend: bool=False) -> Tuple[float, float]:
    """
    Calculate the root mean square of the acceleration and angular velocity components for the specified axes.
    
    Args:
    samples: List of IMU samples.
    axes: String with the axes to calculate the RMS values. Must be a combination of 'x', 'y', and 'z'.
    de_trend: If True, remove the trend from the data.

    Returns:
    Tuple of the root mean square of the acceleration and angular velocity.
    """

    if len(samples) == 0:
        return (0, 0)

    acc_data = []
    gyro_data = []
    if 'x' in axes:
        acc_x = [s.acc_x for s in samples]
        acc_data.append(acc_x)
        gyro_x = [s.gyro_x for s in samples]
        gyro_data.append(gyro_x)
    if 'y' in axes:
        acc_y = [s.acc_y for s in samples]
        acc_data.append(acc_y)
        gyro_y = [s.gyro_y for s in samples]
        gyro_data.append(gyro_y)
    if 'z' in axes:
        acc_z = [s.acc_z for s in samples]
        acc_data.append(acc_z)
        gyro_z = [s.gyro_z for s in samples]
        gyro_data.append(gyro_z)

    rms_acc = rms(tuple(acc_data), de_trend)
    rms_gyro = rms(tuple(gyro_data), de_trend)
    return (rms_acc, rms_gyro)


def running_rms(data: Union[List[float], List[int]], window_size: int, de_trend:bool=False) -> List[float]:
    """
    Calculate the running root mean square of a list of values.
    
    Args:
    data: List of values.
    window_size: Size of the window for the running RMS calculation.
    de_trend: If True, remove the trend from the data.

    Returns:
    List with the running root mean square of the values.
    """
    
    ret = []
    for i in range(len(data)):
        start_idx = max(0, i - window_size)
        stop_idx = min(len(data), i)
        window = data[start_idx:stop_idx]
        if len(window) == 0:
            ret.append(0)
        else:
            ret.append(rms(data[start_idx:stop_idx], de_trend))
    return ret
